/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "AllocArray.h"
#include "Stack.h"
#include "FMapper.h"
#include <stdio.h>
#include <sys/mman.h>
#include <errno.h>
#include "parserExterns.h"

#ifdef USING_MPI
#include <mpi.h>
#include "mpiExterns.h"
#endif

FMapper::FMapper(char* meta) {
  //valid = 0;
  file = 0;
  //arrayf = 0;
  //arrayi = 0;

  yypv.dimset = 0;
  yypv.fnset = 0;
  yypv.fail = 0;
  yypv.skip = 0;
  yypv.min = 0;
  yypv.max = 0;
  yypv.splittotal = 0;

  yyin = fopen(meta, "r");

  if(!yyin) {
    fprintf(stderr, "ERROR: unable to open %s for reading\n\n", meta);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  yyrestart(yyin);
  
  yyparse();

  fclose(yyin);

  if(!yypv.fnset || !yypv.dimset || yypv.fail) {

    /*
    if(yypv.fnset) {
      free(yypv.fn);
    }
    */

    fprintf(stderr, "ERROR: error parsing the meta file %s\n\n", meta);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  dims.SetSize(yypv.stack.Size());
  for(int i = dims.Size() - 1; i >= 0; i = i - 1) {
    dims[i] = yypv.stack.Pop();
    if(dims[i] < 0) {
      yypv.fail = 1;
    }
  }

  if(!yypv.fail) {
    oridims.SetSize(yypv.oristack.Size());  
    for(int i = oridims.Size() - 1; i >= 0; i = i - 1) {
      oridims[i] = yypv.oristack.Pop();
      if(oridims[i] < 0) {
	yypv.fail = 1;
      }
    }
  }

  if(yypv.fail) {
    if(yypv.fnset) {
      free(yypv.fn);
    }

    fprintf(stderr, "ERROR: error parsing the meta file %s\n\n", meta);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  first.SetSize(dims.Size());
  last.SetSize(dims.Size());
  first[0] = 1;
  last[dims.Size() - 1] = 1;
  for(int i = 1; i < dims.Size(); i++) {
    first[i] = first[i - 1] * dims[i - 1];
  }
  for(int i = dims.Size() - 2; i >= 0; i--) {
    last[i] = last[i + 1] * dims[i + 1];
  }

  file = fopen(yypv.fn, "r");
  fn = (char*)malloc(strlen(yypv.fn) * sizeof(char) + 1);
  strcpy(fn, yypv.fn);
  free(yypv.fn);

  if(!file) {
    fprintf(stderr, "ERROR: unable to open %s for reading\n", fn);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  fseek(file, 0, SEEK_END);
  end = ftell(file);
  size = (end - yypv.skip) / 4;
  skip = yypv.skip;
  rewind(file);

  //valid = 1;
  
  /*
  arrayf = (float*)mmap(0, end, PROT_READ, MAP_SHARED, fileno(file), 0);
  arrayi = (int*)arrayf;

  if(arrayf == MAP_FAILED) {
    fprintf(stderr, "\nmmap errno: %d\n", errno);
    fflush(stderr);
    valid = 0;
  }
  */

  min = yypv.min;
  max = yypv.max;
  splittotal = yypv.splittotal;

  /*
  for(int i = 0; i < dims.Size(); i++) {
    dims[i] = (int)pow(2, ceil(log(dims[i]) / log(2)));
  }
  */
}

FMapper::~FMapper() {
  /*
  if(valid) {
    munmap(arrayf, end);
  }
  */

  if(file) {
    fclose(file);
  }

  free(fn);
}

AllocArray<unsigned int>* FMapper::getDims() {
  return &dims;
}

float FMapper::getFloat(AllocArray<unsigned int>& coord) {
  unsigned long offset = 0;
  for(int i = 0; i < first.Size(); i++) {
    offset = offset + coord[i] * first[i];
  }
  //return (arrayf + skip)[offset];
  return getFloat(offset);
}

int FMapper::getInt(AllocArray<unsigned int>& coord) {
  unsigned long offset = 0;
  for(int i = 0; i < last.Size(); i++) {
    offset = offset + coord[i] * last[i];
  }
  //return (arrayi + skip)[offset];
  return getInt(offset);
}

float FMapper::getFloat(unsigned int coord) {
  //return (arrayf + skip)[coord];

  fseek(file, skip + coord * 4, SEEK_SET);
  float f;
  fread(&f, sizeof(float), 1, file);
  return f;
}

int FMapper::getInt(unsigned int coord) {
  //return (arrayi + skip)[coord];

  fseek(file, skip + coord * 4, SEEK_SET);
  int i;
  fread(&i, sizeof(float), 1, file);
  return i;
}

void FMapper::getFloat(float* buffer, unsigned int coord, unsigned int count)
{
  fseek(file, skip + coord * 4, SEEK_SET);
  fread(buffer, sizeof(float), count, file);
}

void FMapper::getInt(int* buffer, unsigned int coord, unsigned int count){
  fseek(file, skip + coord * 4, SEEK_SET);
  fread(buffer, sizeof(float), count, file);
}

void FMapper::getFloat(float* buffer, unsigned int coord, unsigned int count,
		       unsigned int skipcoord)
{
  for(int i = 0; i < count; i = i + 1) {
    fseek(file, skip + coord * 4 + skipcoord * 4 * i, SEEK_SET);
    fread(&buffer[i], sizeof(float), 1, file);
  }
}

void FMapper::getInt(int* buffer, unsigned int coord, unsigned int count,
		     unsigned int skipcoord){
  for(int i = 0; i < count; i = i + 1) {
    fseek(file, skip + coord * 4 + skipcoord * 4 * i, SEEK_SET);
    fread(&buffer[i], sizeof(float), 1, file);
  }
}

unsigned int FMapper::getSize() {
  return size;
}

/*
int FMapper::isValid() {
  return valid;
}
*/

/*
void* FMapper::vector() {
  //return (void*)arrayf;
  return 0;
}
*/

float FMapper::getMin() {
  return min;
}

float FMapper::getMax() {
  return max;
}

void FMapper::close() {
  if(file) {
    fclose(file);
    file = 0;
  }
}

void FMapper::open() {
  if(!file) {
    file = fopen(fn, "r");
    if(!file) {
      fprintf(stderr, "ERROR: unable to open %s for reading", fn);
#ifdef USING_MPI
      MPI_Finalize();
#endif
      exit(1);
    }
  }
}

int FMapper::isOpen() {
  return file != 0;
}

AllocArray<unsigned int>* FMapper::getOriginalDims() {
  return &oridims;
}

unsigned int FMapper::getSplitTotal() {
  return splittotal;
}
