/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "KMeans.h"
#include <stdio.h>
#include <stdlib.h>
#include "clusterOps.h"
#include <string.h>
#include <float.h>
#include "SequentialSpace.h"
#include <math.h>

#ifdef USING_MPI
#include <mpi.h>
#include "mpiExterns.h"
#endif

KMeans::KMeans
(SequentialTimeReader* input, char* outputFormat,
 char* metaExt, char* dataExt, char* infoExt)
{
  this->input = input;
  this->outputFormat = outputFormat;
  this->metaExt = metaExt;
  this->dataExt = dataExt;
  this->infoExt = infoExt;
}

KMeans::~KMeans()
{
}

unsigned int 
KMeans::cluster(unsigned int t, unsigned int k, 
		unsigned int w, unsigned int bins,
		unsigned int vectorType, unsigned int filterType,
		unsigned int distanceType, unsigned int maxiter)
{
  unsigned int l = t - w / 2;
  unsigned int r = t + w / 2 + 1;
  unsigned int ow = w;

  // set up the functions
  void (*getVector) (SequentialTimeReader*, float*, unsigned int,
		     unsigned int, unsigned int);
  void (*findBMU) (float*, double*, unsigned int &, double &,
		   unsigned int, unsigned int);

  if(filterType == BOX_FILTER) {
    if(vectorType == DC_VECTOR) {    
      getVector = &getTimeVectorBox;
    }
    else if(vectorType == DC_DIFF_VECTOR) {
      w = w + w - 1;
      getVector = &getTimeVecAndDiffBox;
    }
    else if(vectorType == DIFF_VECTOR) {
      getVector = &getTimeDiffBox;
    }
    else if(vectorType == HAAR_DC_VECTOR) {
      unsigned int i = w == 1 ? 0 : w / 2 + w % 2;
      while(i > 0) {
	w = w + i;
	i = i == 1 ? 0 : i / 2 + i % 2;
      }

      getVector = &getTimeHaarDCBox;
    }
    else if(vectorType == HAAR_DC_DIFF_VECTOR) {
      unsigned int i = w == 1 ? 0 : w / 2 + w % 2;
      while(i > 0) {
	w = w + i * 2;
	i = i == 1 ? 0 : i / 2 + i % 2;
      }

      getVector = &getTimeHaarDCDiffBox;
    }
    else if(vectorType == HAAR_DIFF_VECTOR) {
      unsigned int i = w == 1 ? 0 : w / 2 + w % 2;
      w = 0;
      while(i > 0) {
	w = w + i;
	i = i == 1 ? 0 : i / 2 + i % 2;
      }

      getVector = &getTimeHaarDiffBox;
    }
  }

  if(distanceType == EUCLIDEAN_DISTANCE) {
    findBMU = &findBMUL2;
  }
    
  // pointers to data
  double* sums[2];
  float* errors[2];
  unsigned int* counts[2];
  float totalerror[2];
  
  // keep track of the last few states, so to not toggle
  // between centroid point swapping and can terminate
  unsigned int laststate = 0;
#define HASH_STATES 32
  unsigned int states[HASH_STATES];
  bzero(states, sizeof(unsigned int) * HASH_STATES);

  // data
  double bufSums[2 * k * w];
  float bufErrors[2 * k];
  unsigned int bufCounts[2 * k];  
#ifdef USING_MPI
  float dest_float[k * w];
  double dest_double[k * w];
  unsigned int dest_uint[w * k * bins];
#endif

  // set up pointers
  for(int i = 0; i < 2; i = i + 1) {
    sums[i] = &bufSums[i * k * w];
    errors[i] = &bufErrors[i * k];
    counts[i] = &bufCounts[i * k];
  }

  // indices
  unsigned int last = 0;
  unsigned int work = 1;

  // get the min-max for this set
  float min = FLT_MAX;
  float max = FLT_MIN;

  float cmin[w];
  float cmax[w];

  for(unsigned int j = 0; j < w; j = j + 1) {
    cmin[j] = FLT_MAX;
    cmax[j] = FLT_MIN;
  }

  for(unsigned int i = 0; i < input->getSize(); i = i + 1) {
    float buffer[w];
    getVector(input, buffer, i, l, r);

    for(unsigned int j = 0; j < w; j = j + 1) {
      cmin[j] = cmin[j] > buffer[j] ? buffer[j] : cmin[j];
      cmax[j] = cmax[j] < buffer[j] ? buffer[j] : cmax[j];

      min = min > buffer[j] ? buffer[j] : min;
      max = max < buffer[j] ? buffer[j] : max;
    }
  }

#ifdef USING_MPI
  MPI_Allreduce(&min, dest_float, 1, MPI_FLOAT, 
		MPI_MIN, MPI_COMM_WORLD);
  min = dest_float[0];

  MPI_Allreduce(&max, dest_float, 1, MPI_FLOAT, 
		MPI_MAX, MPI_COMM_WORLD);
  max = dest_float[0];

  MPI_Allreduce(cmin, dest_float, w, MPI_FLOAT, 
		MPI_MIN, MPI_COMM_WORLD);
  memcpy(cmin, dest_float, sizeof(float) * w);
  
  MPI_Allreduce(cmax, dest_float, w, MPI_FLOAT, 
		MPI_MAX, MPI_COMM_WORLD);
  memcpy(cmax, dest_float, sizeof(float) * w);
#endif
  double cscale[w];

  for(unsigned int j = 0; j < w; j = j + 1) {
    cscale[j] = bins / (cmax[j] - cmin[j]);
  }

  // try and get clusters from time histogram
  unsigned int histograms[w][k][bins];
  bzero(histograms, sizeof(unsigned int) * w * k * bins);
  
  for(unsigned int i = 0; i < input->getSize(); i = i + 1) {
    float buffer[w];
    getVector(input, buffer, i, l, r);
    
    for(int j = 0; j < w; j = j + 1) {
      unsigned int index = (unsigned int)((buffer[j] - cmin[j]) * cscale[j]);
      index = index >= bins ? bins - 1 : index;
      histograms[j][0][index] = histograms[j][0][index] + 1;
    }
  }
  
#ifdef USING_MPI
  MPI_Allreduce(histograms, dest_uint, w * bins * k, MPI_UNSIGNED, 
		MPI_SUM, MPI_COMM_WORLD);
  memcpy(histograms, dest_uint, sizeof(unsigned int) * w * bins * k);
#endif
  
  // use the cdf to find some centroids
  for(unsigned int j = 0; j < w; j = j + 1) {
    unsigned int cdf[bins];
    cdf[0] = histograms[j][0][0];
    for(unsigned int i = 1; i < bins; i = i + 1) {
      cdf[i] = cdf[i - 1] + histograms[j][0][i];
    }
    
    unsigned int v = 0; 
    for(unsigned int i = 0; i < k; i = i + 1) {
      float stop = (float)(i * 4 + 2) * cdf[bins - 1] / (k * 4);
      
      while(cdf[v] < stop && v < bins - 1) {
	v = v + 1;
      }
      
      float fv = v;
      if(v == 0) {
	fv = stop / cdf[v];
      }
      else {
	fv = (stop - cdf[v - 1]) / (cdf[v] - cdf[v - 1]);
      }
      
      // set the centroid
      sums[last][i * w + j] = (v + fv) / cscale[j] + cmin[j];
    }
  }
	
#ifdef USING_MPI
  if(mpi_rank == 0) {
#endif
    for(unsigned int i = 0; i < k; i = i + 1) {
      for(unsigned int j = 0; j < w; j = j + 1) {
	printf("%f ", sums[last][i * w + j]);
      }
      printf("\n");
    }
#ifdef USING_MPI
  }
#endif
  
  unsigned int iter = 0;
  unsigned int changed = 1;

  // while the centroids haven't changed
  while(changed) {
    clearBuffer(sums[work], errors[work], counts[work], w, k);
    
    // assign each vector to a BMU centroid
    for(int i = 0; i < input->getSize(); i = i + 1) {
      float buffer[w];
      getVector(input, buffer, i, l, r);
      
      unsigned int bmu;
      double mindist;
      
      // find BMU
      findBMU(buffer, sums[last], bmu, mindist, w, k);

      // add to BMU
      sum(&sums[work][bmu * w], buffer, w);
      counts[work][bmu] = counts[work][bmu] + 1;
    }

#ifdef USING_MPI
    // sum the results
    MPI_Allreduce(sums[work], dest_double, k * w, MPI_DOUBLE, MPI_SUM,
		  MPI_COMM_WORLD);
    MPI_Allreduce(counts[work], dest_uint, k, MPI_UNSIGNED, MPI_SUM,
		  MPI_COMM_WORLD);
    
    memcpy(sums[work], dest_double, k * w * sizeof(double));
    memcpy(counts[work], dest_uint, k * sizeof(unsigned int));
#endif

    // calculate new centroid
    averageSums(sums[work], counts[work], w, k);

    if(iter >= maxiter) {
      changed = 0;
    }
    else {
      // detect if repeated clustering using hash
      laststate = (laststate + 1) % HASH_STATES;
      states[laststate] = jenkins(sums[work], counts[work], w, k);
      for(int i = 0; i < HASH_STATES; i = i + 1) {
	if(i != laststate && states[laststate] == states[i]) {
	  changed = 0;
	  break;
	}
      }
    }

    // rotate buffers
    last = work;
    work = (work + 1) % 2;

    // show update
    iter = iter + 1;
#ifdef USING_MPI
    if(mpi_rank == 0) {
#endif
      printf("\ri: %d", iter);
    fflush(stdout);
#ifdef USING_MPI
    }
#endif
  } // while not changed

  // rotate back to use last again, to repeat the same work
  last = work;
  work = (work + 1) % 2;

  // write out the data
  SequentialSpaceWriter output
    (outputFormat, metaExt, dataExt, 1, *input->getOriginalDims(), t, 1);
  clearBuffer(sums[work], errors[work], counts[work], w, k);

  // clear the histograms
  bzero(histograms, sizeof(unsigned int) * w * k * bins);
  double scale = bins / (max - min);
  
  // assign each vector to a BMU centroid
  for(int i = 0; i < input->getSize(); i = i + 1) {
    float buffer[w];
    getVector(input, buffer, i, l, r);
 
    unsigned int bmu;
    double mindist;
    
    // find BMU
    findBMU(buffer, sums[last], bmu, mindist, w, k);

    // write to output the index
    float temp = bmu;
    output.writeTimeVector(&temp);
    
    // add the histogram count
    for(int j = 0; j < w; j = j + 1) {
      unsigned int index = (unsigned int)((buffer[j] - min) * scale);
      index = index >= bins ? bins - 1 : index;
      histograms[j][bmu][index] = histograms[j][bmu][index] + 1;
    }

    // add to BMU
    sum(&sums[work][bmu * w], buffer, w);
    errors[work][bmu] = errors[work][bmu] + (float)mindist;
    counts[work][bmu] = counts[work][bmu] + 1;
  }

#ifdef USING_MPI
  // sum the results
  MPI_Allreduce(sums[work], dest_double, k * w, MPI_DOUBLE, MPI_SUM,
		MPI_COMM_WORLD);
  MPI_Allreduce(counts[work], dest_uint, k, MPI_UNSIGNED, MPI_SUM,
		MPI_COMM_WORLD);
  
  memcpy(sums[work], dest_double, k * w * sizeof(double));
  memcpy(counts[work], dest_uint, k * sizeof(unsigned int));
  
  // sum the results
  MPI_Allreduce(errors[work], dest_float, k, MPI_FLOAT, MPI_SUM,
		MPI_COMM_WORLD);
  MPI_Allreduce(histograms, dest_uint, bins * w * k, MPI_UNSIGNED, MPI_SUM,
		MPI_COMM_WORLD);
  
  memcpy(errors[work], dest_float, k * sizeof(float));
  memcpy(histograms, dest_uint, bins * w * k * sizeof(unsigned int));
#endif
  
  // calculate new centroid
  averageSums(sums[work], counts[work], w, k);
  totalerror[work] = averageErrors(errors[work], counts[work], k);

  // report the clusters
#ifdef USING_MPI
  if(mpi_rank == 0) {
#endif
  printf("\n");
  for(int i = 0; i < k; i = i + 1) {
    for(int j = 0; j < w; j = j + 1) {
      printf("%f ", sums[work][i * w + j]);
    }
    printf("\nE: %e C: %u\n", errors[work][i], counts[work][i]);
  }
  printf("AE: %e\n\n", totalerror[work]);
#ifdef USING_MPI
  }
#endif

  // write the information file
#ifdef USING_MPI
  if(mpi_rank == 0)
#endif
  writeInfo(outputFormat, infoExt, t, 
	    sums[work], errors[work], counts[work], (unsigned int *)histograms,
	    KMEANS_CLUSTERING, vectorType, filterType,  distanceType,
	    w, k, bins, min, max, iter, totalerror[work]);

  // return the k
  return k;
}
