/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "FMapper.h"
#include "MMapper.h"

#include "SequentialSpace.h"
#include <stdlib.h>

SequentialSpaceReader::SequentialSpaceReader
(int mmap, char* format, 
unsigned int start, unsigned int end, unsigned int skip)
{
  tc = (end - start) / skip;
  mappers = new Mapper* [tc];
  
  char temp[strlen(format) + 256];

  // open all the mappers and close them
  for(int i = 0; i < tc; i = i + 1) {
    sprintf(temp, format, i * skip + start);

    if(mmap) {
      mappers[i] = new MMapper(temp);
    }
    else {
      mappers[i] = new FMapper(temp);
    }

    mappers[i]->close();
  }

  // open the first
  current = 0;
  mappers[0]->open();

  vc = mappers[0]->getSize();

  dims.SetSize(mappers[0]->getDims()->Size());
  for(int i = 0; i < dims.Size(); i = i + 1) {
    dims[i] = (*mappers[0]->getDims())[i];
  }
}

SequentialSpaceReader::~SequentialSpaceReader()
{
  for(unsigned int i = 0; i < tc; i = i + 1) {
    // TO FIX: why do I have to do a close? the deconstructor should close it
    // TO FIX: leaves a free handle and file handle memory leak with a
    // TO FIX: unlink after a delete
    mappers[i]->close();
    delete mappers[i];
  }
  
  delete [] mappers;
}

float SequentialSpaceReader::getFloat(unsigned int coord) {
  return mappers[current]->getFloat(coord);
}

void SequentialSpaceReader::setTime(unsigned int time) {
  mappers[current]->close();
  current = time;
  mappers[current]->open();
}

unsigned int SequentialSpaceReader::getSize() {
  return vc;
}

unsigned int SequentialSpaceReader::getMaxTime() {
  return tc;
}

unsigned int SequentialSpaceReader::getTime() {
  return current;
}

AllocArray<unsigned int>* SequentialSpaceReader::getDims() {
  return &dims;
}

void SequentialSpaceReader::getFloat
(float* buffer, unsigned int coord, unsigned int count)
{
  mappers[current]->getFloat(buffer, coord, count);
}

SequentialSpaceWriter::SequentialSpaceWriter
(char* format, char* meta, char* data, unsigned int tc,
 AllocArray<unsigned int>& oridims, unsigned int offset, unsigned int skip)
{
  this->tc = tc;
  this->offset = offset;
  this->skip = skip;

    // set up the file names
  metafn = new char[strlen(format) + strlen(meta) + 1];
  datafn = new char[strlen(format) + strlen(data) + 1];

  sprintf(metafn, "%s%s", format, meta);
  sprintf(datafn, "%s%s", format, data);

  dims.SetSize(oridims.Size());
  for(int i = 0; i < dims.Size(); i = i + 1) {
    dims[i] = oridims[i];
  }

  char temp[strlen(datafn) + 256];
  outputs = new FILE*[tc];

  // open the outputs
  for(unsigned int i = 0; i < tc; i = i + 1) {
    sprintf(temp, datafn, i * skip + offset);
    outputs[i] = fopen(temp, "w");

    if(!outputs[i]) {
      fprintf(stderr, "ERROR: unable to open %s for writing\n\n", temp);
      exit(1);
    }
  }
}

SequentialSpaceWriter::~SequentialSpaceWriter()
{
  if(outputs) {
    close();
  }

  delete [] metafn;
  delete [] datafn;
}

void SequentialSpaceWriter::close()
{
  char temp[strlen(datafn) + 256];
  FILE* output;

  for(unsigned int i = 0; i < tc; i = i + 1) {
    fclose(outputs[i]);

    sprintf(temp, metafn, i * skip + offset);
    output = fopen(temp, "w");

    if(!output) {
      fprintf(stderr, "ERROR: unable to open %s for writing\n\n", temp);
      exit(1);
    }

    fprintf(output, "dimensions:");
    for(int j = 0; j < dims.Size(); j = j + 1) {
      fprintf(output, " %d", dims[j]);
    }
    fprintf(output, "\nfilename: ");
    fprintf(output, datafn, i * skip + offset);
    fprintf(output, "\n");

    fclose(output);
  }

  delete [] outputs;
  outputs = 0;
}  
    
void SequentialSpaceWriter::writeTimeVector(float* buffer)
{
  for(unsigned int i = 0; i < tc; i = i + 1) {
    fwrite(&(buffer[i]), sizeof(float), 1, outputs[i]);
  }
}
