/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "MMapper.h"
#include "FMapper.h"

#include "SequentialTime.h"
#include <stdlib.h>

#ifdef USING_MPI
#include <mpi.h>
#include "mpiExterns.h"
#endif

SequentialTimeWriter::SequentialTimeWriter
(char* format, char* meta, char* data, 
 unsigned int tc, unsigned int max, AllocArray<unsigned int>& dims,
 unsigned int splittotal)
{
  this->max = max / sizeof(float) / tc;
  this->tc = tc;
  this->vc = splittotal;

  minset = 0;
  maxset = 0;

  if(this->vc == 0) {
    fprintf(stderr, "ERROR: incorrect splittotal of %u\n\n", this->vc);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  // set up the file names
  metafn = new char[strlen(format) + strlen(meta) + 1];
  datafn = new char[strlen(format) + strlen(data) + 1];

  sprintf(metafn, "%s%s", format, meta);
  sprintf(datafn, "%s%s", format, data);

  count = 0;
  current = 0;

  oridims.SetSize(dims.Size());
  for(int i = 0; i < oridims.Size(); i = i + 1) {
    oridims[i] = dims[i];
  }

  char temp[strlen(datafn) + 256];

  // open the first output
  sprintf(temp, datafn, 0);
  output = fopen(temp, "w");

  if(!output) {
    fprintf(stderr, "ERROR: unable to open %s for writing\n\n", temp);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }
}

SequentialTimeWriter::~SequentialTimeWriter()
{
  close();

  delete [] metafn;
  delete [] datafn; 
}

void SequentialTimeWriter::writeTimeVector(float* buffer)
{
  // if it exceeds the max, open a new one
  if(count >= max) {
    close();
    count = 0;
    current = current + 1;

    char temp[strlen(datafn) + 256];
    sprintf(temp, datafn, current);
    output = fopen(temp, "w");

    if(!output) {
      fprintf(stderr, "ERROR: unable to open %s for writing\n\n", temp);
#ifdef USING_MPI
      MPI_Finalize();
#endif
      exit(1);
    }
  }

  // write the data
  count = count + 1;
  fwrite(buffer, sizeof(float), tc, output);
}

void SequentialTimeWriter::close() 
{
  if(count > 0) {
    fclose(output);

    char temp[strlen(metafn) + 256];

    // write the meta data
    sprintf(temp, metafn, current);
    output = fopen(temp, "w");

    if(!output) {
      fprintf(stderr, "ERROR: unable to open %s for writing\n\n", temp);
#ifdef USING_MPI
      MPI_Finalize();
#endif
      exit(1);
    }

    fprintf(output, "dimensions: %u %u\nfilename: ", tc, count);
    fprintf(output, datafn, current);
    fprintf(output, "\n");
    if(oridims.Size() > 0) {
      fprintf(output, "original:");
      for(int i = 0; i < oridims.Size(); i = i + 1) {
	fprintf(output, " %u", oridims[i]);
      }
      fprintf(output, "\n");
    }
    fprintf(output, "splittotal: %u\n", vc);
    if(minset) {
      fprintf(output, "min: %e\n", minfloat);
    }
    if(maxset) {
      fprintf(output, "max: %e\n", maxfloat);
    }

    fclose(output);
  }
}

void SequentialTimeWriter::setMin(float m) {
  minfloat = m;
  minset = 1;
}

void SequentialTimeWriter::setMax(float m) {
  maxfloat = m;
  maxset = 1;
}

SequentialTimeReader::SequentialTimeReader(int mmap, char* format)
{
  char temp[strlen(format) + 256];
  sprintf(temp, format, 0);

  FMapper* first = new FMapper(temp);

  tc = (*(first->getDims()))[0];
  vc = first->getSplitTotal();

  if(vc == 0) {
    fprintf(stderr, "ERROR: incorrect splittotal of %u\n\n", vc);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  oridims.SetSize(first->getOriginalDims()->Size());
  for(int i = 0; i < oridims.Size(); i = i + 1) {
    oridims[i] = (*(first->getOriginalDims()))[i];
  }

  chunk = (*(first->getDims()))[1];
  fc = vc / chunk + (vc % chunk > 0 ? 1 : 0);

  delete first;

  mappers = new Mapper*[fc];

  // open all the mappers and close them
  for(int i = 0; i < fc; i = i + 1) {
    sprintf(temp, format, i);
    
    if(mmap) {
      mappers[i] = new MMapper(temp);
    }
    else {
      mappers[i] = new FMapper(temp);
    }

    // get the min/max from the last file
    if(i == fc - 1) {
      minfloat = mappers[i]->getMin();
      maxfloat = mappers[i]->getMax();
    }

    mappers[i]->close();
  }

  // open the first
  current = 0;
  mappers[0]->open();
}

SequentialTimeReader::~SequentialTimeReader()
{
  for(int i = 0; i < fc; i = i + 1) {
    delete mappers[i];
  }
  
  delete [] mappers;
}

unsigned int SequentialTimeReader::getMaxTime()
{
  return tc;
}

unsigned int SequentialTimeReader::getSize()
{
  return vc;
}

AllocArray<unsigned int>* SequentialTimeReader::getOriginalDims()
{
  return &oridims;
}

void SequentialTimeReader::getTimeVector
(float *buffer, unsigned int coord)
{
  unsigned int f = coord / chunk;
  unsigned int c = coord % chunk;
  if(f != current) {
    mappers[current]->close();
    mappers[f]->open();
    current = f;
  }

  mappers[f]->getFloat(buffer, c * tc, tc);
}

void SequentialTimeReader::getTimeVector
(float *buffer, unsigned int coord, unsigned int start, unsigned int end)
{
  unsigned int f = coord / chunk;
  unsigned int c = coord % chunk;
  if(f != current) {
    mappers[current]->close();
    mappers[f]->open();
    current = f;
  }

  mappers[f]->getFloat(buffer, c * tc + start, end - start);
}

void SequentialTimeReader::getTimeVector
(float *buffer, unsigned int coord, 
 unsigned int start, unsigned int end, unsigned int skip)
{
  unsigned int f = coord / chunk;
  unsigned int c = coord % chunk;
  if(f != current) {
    mappers[current]->close();
    mappers[f]->open();
    current = f;
  }

  mappers[f]->getFloat(buffer, c * tc + start, end - start, skip);
}

float SequentialTimeReader::getMin() {
  return minfloat;
}

float SequentialTimeReader::getMax() {
  return maxfloat;
}
