/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "clusterCommand.h"
#include <stdio.h>
#include <stdlib.h>

#include "SequentialSpace.h"
#include "SequentialTime.h"

#include "clusterOps.h"
#include "KMeans.h"

#ifdef USING_MPI
#include <mpi.h>
#include "mpiExterns.h"
#endif

int main(int argc, char* argv[])
{
#ifdef USING_MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
#endif

  CommandArguments* args = commandParse(argc, argv);

#ifdef USING_MPI
  char inputformat[strlen(args->inputFormat) + 256];
  sprintf(inputformat, "%s[%d]", args->inputFormat, mpi_rank); 

  SequentialTimeReader* input =
    new SequentialTimeReader(args->useMMap, inputformat);
#else
  SequentialTimeReader* input =
    new SequentialTimeReader(args->useMMap, args->inputFormat);
#endif

  // make sure it is in bounds
  if(args->startFile < args->timeWidth / 2) {
    args->startFile = args->timeWidth / 2;
  }

  if(args->endFile + args->timeWidth / 2 > input->getMaxTime()) {
    args->endFile = input->getMaxTime() - args->timeWidth / 2;
  }

  if(args->startFile >= args->endFile) {
    args->endFile = input->getMaxTime() - args->timeWidth / 2;
  }

  // use K-means
  if(args->clusteringType == KMEANS_CLUSTERING) {

#ifdef USING_MPI
    char metaformat[strlen(args->metaExt) + 256];
    sprintf(metaformat, "%s[%d]", args->metaExt, mpi_rank); 
    char dataformat[strlen(args->dataExt) + 256];
    sprintf(dataformat, "%s[%d]", args->dataExt, mpi_rank); 

    KMeans cluster(input, args->outputFormat, 
		   metaformat, dataformat, args->infoExt);
#else
    KMeans cluster(input, args->outputFormat, 
		   args->metaExt, args->dataExt, args->infoExt);
#endif

    // cluster each time step
    for(int i = args->startFile; i < args->endFile; i = i + args->skipFile) {
#ifdef USING_MPI
      if(mpi_rank == 0)
#endif
      printf("t: %d\n", i);
      
      // cluster
      cluster.cluster
	(i, args->numClusters, args->timeWidth, args->histogramBins, 
	 args->vectorType, args->filterType, args->distanceType,
	 args->maxiter);

#ifdef USING_MPI
      // wait for writes
      MPI_Barrier(MPI_COMM_WORLD);

      // sew the data back together using 0
      if(mpi_rank == 0) {
	SequentialSpaceWriter 
	  sewn(args->outputFormat, args->metaExt, args->dataExt,
	       1, *input->getOriginalDims(), i, 0);

	for(int j = 0; j < mpi_size; j = j + 1) {

	  char brickmeta[strlen(args->outputFormat) +
			 strlen(args->metaExt) + 256];
	  sprintf(brickmeta, "%s%s[%d]", 
		  args->outputFormat, args->metaExt, j);

	  char brickdata[strlen(args->outputFormat) +
			 strlen(args->dataExt) + 256];
	  sprintf(brickdata, "%s%s[%d]", 
		  args->outputFormat, args->dataExt, j);

	  // read the data and write to the new file
	  SequentialSpaceReader* brick = new
	    SequentialSpaceReader(args->useMMap, brickmeta, i, i + 1, 1);

	  for(unsigned int m = 0; m < brick->getSize(); m = m + 1) {
	    float val;
	    val = brick->getFloat(m);
	    sewn.writeTimeVector(&val);
	  }

	  delete brick;

	  char tempfn[strlen(brickmeta) + 256];
	  sprintf(tempfn, brickmeta, i);
	  remove(tempfn);
	  sprintf(tempfn, brickdata, i);
	  remove(tempfn);
	}
      }

      // wait for 0
      MPI_Barrier(MPI_COMM_WORLD);
#endif
    }
  }

  delete input;
  delete args;

#ifdef USING_MPI
  MPI_Finalize();
#endif
}
