/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "clusterOps.h"
#include <strings.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>

#ifdef USING_MPI
#include <mpi.h>
#include "mpiExterns.h"
#endif

double distL2(float* a, double* b, unsigned int w) {
  double dist = 0;
  for(int i = 0; i < w; i = i + 1) {
    double val = a[i] - b[i];
    dist = dist + val * val;
  }
  return dist;
}

/*
float distDTW2(float *a, float* b, unsigned int w) 
{
  float DTW[w + 1][w + 1];
  
  for(unsigned int i = 1; i <= w; i = i + 1) {
    DTW[0][i] = FLT_MAX;
    DTW[i][0] = FLT_MAX;
  }
  DTW[0][0] = 0;

  for(unsigned int i = 1; i <= w; i = i + 1) {
    for(unsigned int j = 1; j <= w; j = j + 1) {
      float cost = a[i] - b[j];
      cost = cost * cost;

      float x = DTW[i - 1][j];
      float y = DTW[i][j - 1];
      float z = DTW[i - 1][j - 1];

      DTW[i][j] = x < y ? (x < z ? x : z) : (y < z ? y : z);
      DTW[i][j] = DTW[i][j] + cost;
    }
  }

  return DTW[w][w];
}
*/

void sum(double* sum, float* a, unsigned int w) {
  for(int i = 0; i < w; i = i + 1) {
    sum[i] = sum[i] + a[i];
  }
}

void clearBuffer(double* sums, float* errors, unsigned int* counts, 
		 unsigned int w, unsigned int k) {
  bzero(sums, sizeof(double) * k * w);
  bzero(errors, sizeof(float) * k);
  bzero(counts, sizeof(unsigned int) * k);
}

void findBMUL2(float* buffer, double* sums,
	       unsigned int &bmu, double &mindist,
	       unsigned int w, unsigned int k) {
  // find the BMU
  bmu = 0;
  mindist = distL2(buffer, &sums[0], w);

  for(int i = 1; i < k; i = i + 1) {
    double dist = distL2(buffer, &sums[i * w], w);
    if(dist < mindist) {
      bmu = i;
      mindist = dist;
    }
  }
}

/*
void findBMUDTW2(float* buffer, float* sums,
		 unsigned int &bmu, float &mindist,
		 unsigned int w, unsigned int k) {
  // find the BMU
  bmu = 0;
  mindist = distDTW2(buffer, &sums[0], w);

  for(int i = 1; i < k; i = i + 1) {
    float dist = distL2(buffer, &sums[i * w], w);
    if(dist < mindist) {
      bmu = i;
      mindist = dist;
    }
  }
}
*/

void averageSums(double* sums, unsigned int* counts,
		 unsigned int w, unsigned int k)
{
  for(int i = 0; i < k; i = i + 1) {
    if(counts[i] > 0) {
      double divisor = 1.0 / (double)counts[i];
      for(int j = 0; j < w; j = j + 1) {
	sums[i * w + j] = sums[i * w + j] * divisor;
      }
    }
  }
}

float averageErrors(float* errors, unsigned int* counts, unsigned int k)
{
  float totalerror = 0;
  unsigned int totalcount = 0;

  for(int i = 0; i < k; i = i + 1) {
    totalerror = totalerror + errors[i];
    totalcount = totalcount + counts[i];
    if(counts[i] > 0) {
      errors[i] = errors[i] / counts[i];
    }
  }

  return totalerror / totalcount;
}

unsigned int 
jenkins(double* sums, unsigned int* counts,
	unsigned int w, unsigned int k)
{
  unsigned int hash = 0;

  for(int i = 0; i < k; i = i + 1) {
    for(int j = 0; j < w; j = j + 1) {
      double temp = sums[i * w + j];
      for(int m = 0; m < sizeof(double); m = m + 8) {
	hash += (*((unsigned int*)&temp) >> m) & 0xff;
	hash += (hash << 10);
	hash ^= (hash >> 6);
      }
    }
    for(int m = 0; m < sizeof(unsigned int); m = m + 8) {
      /*
      a = (a + ((*((unsigned int*)&errors[i]) 
		 >> m) & 0xff)) % 65521;
      b = (b + a) % 65521;
      */
      hash += (counts[i] >> m) & 0xff;
      hash += (hash << 10);
      hash ^= (hash >> 6);
    }
  }

  hash += (hash << 3);
  hash ^= (hash >> 11);
  hash += (hash << 15);

  return hash;
}

void calcDifferences(float* buffer, unsigned int b) {
  for(int i = 0; i < b - 1; i = i + 1) {
    buffer[b + i] = buffer[i + 1] - buffer[i];
  }
}

void calcDifferences(float* buffer, float* dest, unsigned int b) {
  for(int i = 0; i < b - 1; i = i + 1) {
    dest[i] = buffer[i + 1] - buffer[i];
  }
  dest[b - 1] = 0;
}

void getTimeVectorBox(SequentialTimeReader* input, float* buffer,
		      unsigned int coord, unsigned int l, unsigned int r)
{
  input->getTimeVector(buffer, coord, l, r);
}

void getTimeVecAndDiffBox(SequentialTimeReader* input, float* buffer,
			  unsigned int coord, unsigned int l, unsigned int r)
{
  input->getTimeVector(buffer, coord, l, r);
  calcDifferences(buffer, r - l);
}

void getTimeDiffBox(SequentialTimeReader* input, float* buffer,
		    unsigned int coord, unsigned int l, unsigned int r)
{
  float temp[r - l];
  input->getTimeVector(temp, coord, l, r);
  calcDifferences(temp, buffer, r - l);
}

void getTimeHaarDCBox(SequentialTimeReader* input, float* buffer,
		      unsigned int coord, unsigned int l, unsigned int r)
{
  input->getTimeVector(buffer, coord, l, r);
  calcHaarDC(buffer, r - l);
}

void getTimeHaarDCDiffBox(SequentialTimeReader* input, float* buffer,
			  unsigned int coord, unsigned int l, unsigned int r)
{
  input->getTimeVector(buffer, coord, l, r);
  calcHaarDCDiff(buffer, r - l);
}

void getTimeHaarDiffBox(SequentialTimeReader* input, float* buffer,
			  unsigned int coord, unsigned int l, unsigned int r)
{
  float temp[r - l];
  input->getTimeVector(temp, coord, l, r);
  calcHaarDiff(temp, buffer, r - l);
}

void calcHaarDC(float* buffer, unsigned int b) {
  unsigned int m = 0;
  unsigned int n = b;
  unsigned int r = b;
  float avg = 0.0;
  for(unsigned int i = 0; i < b; i = i + 1) {
    avg = avg + buffer[i];
  }
  avg = avg / b;
  b = b == 1 ? 0 : b / 2 + (b % 2);
  while(b > 0) {
    for(unsigned int i = 0; i < b; i = i + 1) {
      if(i * 2 + 1 < r) {
	buffer[n + i] = (buffer[m + i * 2] + buffer[m + i * 2 + 1]) * .5;
      }
      else {
	buffer[n + i] = (buffer[m + i * 2] + avg) * .5;
      }
    }

    m = n;
    n = n + b;
    r = b;
    b = b == 1 ? 0 : b / 2 + (b % 2);
  }
}

void calcHaarDCDiff(float* buffer, unsigned int b) {
  unsigned int m = 0;
  unsigned int n = b;
  unsigned int r = b;
  float avg = 0.0;
  for(unsigned int i = 0; i < b; i = i + 1) {
    avg = avg + buffer[i];
  }
  avg = avg / b;
  b = b == 1 ? 0 : b / 2 + (b % 2);
  while(b > 0) {
    for(unsigned int i = 0; i < b; i = i + 1) {
      if(i * 2 + 1 < r) {
	buffer[n + i + b] = buffer[m + i * 2 + 1] - buffer[m + i * 2];
	buffer[n + i] = (buffer[m + i * 2] + buffer[m + i * 2 + 1]) * .5;
      }
      else {
	buffer[n + i + b] = avg - buffer[m + i * 2];
	buffer[n + i] = (buffer[m + i * 2] + avg) * .5;
      }
    }

    m = n;
    n = n + b * 2;
    r = b;
    b = b == 1 ? 0 : b / 2 + (b % 2);
  }
}

void calcHaarDiff(float* buffer, float* dest, unsigned int b) {
  unsigned int n = 0;
  unsigned int r = b;
  float avg = 0.0;
  for(unsigned int i = 0; i < b; i = i + 1) {
    avg = avg + buffer[i];
  }
  avg = avg / b;
  b = b == 1 ? 0 : b / 2 + (b % 2);
  while(b > 0) {
    for(unsigned int i = 0; i < b; i = i + 1) {
      if(i * 2 + 1 < r) {
	dest[n + i] = buffer[i * 2 + 1] - buffer[i * 2];
	buffer[i] = (buffer[i * 2] + buffer[i * 2 + 1]) * .5;
      }
      else {
	dest[n + i] = avg - buffer[i * 2];
	buffer[i] = (buffer[i * 2] + avg) * .5;
      }
    }

    n = n + b;
    r = b;
    b = b == 1 ? 0 : b / 2 + (b % 2);
  }
}

void learn(double* sum, float* buffer, unsigned int w, float rate) {
  for(int i = 0; i < w; i = i + 1) {
    sum[i] = sum[i] + rate * (buffer[i] - sum[i]);
  }
}

void writeInfo(char* format, char* infoExt, unsigned int t,
	       double* sums, float* errors, unsigned int* counts,
	       unsigned int* histograms,
	       unsigned int clusteringType, unsigned int vectorType,
	       unsigned int filterType, unsigned int distanceType,
	       unsigned int w, unsigned int k,
	       unsigned int bins, float min, float max, unsigned int iter,
               float aveerr)
{
  char temp[strlen(format) + strlen(infoExt) + 1];
  sprintf(temp, "%s%s", format, infoExt);
  char fn[strlen(temp) + 256];
  sprintf(fn, temp, t);

  FILE* output = fopen(fn, "w");
  if(!output) {
    fprintf(stderr, "ERROR: unable to open %s for writing\n\n", fn);
#ifdef USING_MPI
    MPI_Finalize();
#endif
    exit(1);
  }

  fprintf(output, "clusteringtype: ");
  switch(clusteringType) {
  case KMEANS_CLUSTERING:
    fprintf(output, "KMEANS\n");
    break;
  default:
    fprintf(output, "\n");
  }

  fprintf(output, "vectortype: ");
  switch(vectorType) {
  case DC_VECTOR:
    fprintf(output, "DC\n");
    break;
  case DC_DIFF_VECTOR:
    fprintf(output, "DCDIFF\n");
    break;
  case DIFF_VECTOR:
    fprintf(output, "DIFF\n");
    break;
  case HAAR_DC_VECTOR:
    fprintf(output, "HAARDC\n");
    break;
  case HAAR_DC_DIFF_VECTOR:
    fprintf(output, "HAARDCDIFF\n");
    break;
  case HAAR_DIFF_VECTOR:
    fprintf(output, "HAARDIFF\n");
    break;
  default:
    fprintf(output, "\n");
  }

  fprintf(output, "filtertype: ");
  switch(filterType) {
  case BOX_FILTER:
    fprintf(output, "BOX\n");
    break;
  default:
    fprintf(output, "\n");
  }

  fprintf(output, "distancetype: ");
  switch(distanceType) {
  case EUCLIDEAN_DISTANCE:
    fprintf(output, "L2\n");
    break;
  default:
    fprintf(output, "\n");
  }

  fprintf(output, "vectorwidth: %u\n", w);
  fprintf(output, "histogrambins: %u\n", bins);
  fprintf(output, "min: %e\n", min);
  fprintf(output, "max: %e\n", max);

  fprintf(output, "timestep: %u\n", t);
  fprintf(output, "clusters: %u\n", k);
  fprintf(output, "iterations: %u\n", iter);
  fprintf(output, "averageerror: %e\n", aveerr);

  for(int i = 0; i < k; i = i + 1) {
    fprintf(output, "cluster:");
    for(int j = 0; j < w; j = j + 1) {
      fprintf(output, " %e", sums[i * w + j]);
    }
    fprintf(output, "\n");
    fprintf(output, "error: %e\n", errors[i]);
    fprintf(output, "count: %u\n", counts[i]);
    for(int j = 0; j < w; j = j + 1) {
      fprintf(output, "timehistogram: ");
      for(int m = 0; m < bins; m = m + 1) {
	fprintf(output, "%u ", histograms[j * k * bins + i * bins + m]);
      }
      fprintf(output, "\n");
    }
  }

  fclose(output);
}
