/*
 * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of The Ohio State University, Jonathan
 *       L. Woodring, Han-Wei Shen, nor the names of its contributors
 *       may be used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
 * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
 * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "psplitCommand.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <mpi.h>
#include "mpiExterns.h"

void setDefaults(CommandArguments* args) {
  args->startFile = 0;
  args->endFile = 0;
  args->skipFile = 1;
  args->useMMap = 1;
  args->maxFile = 1024 * 1024 * 1024;
  
  args->inputFormat = 0;
  args->outputFormat = 0;
  
  args->metaExt = ".ptmeta";
  args->dataExt = ".ptdata";
}

char* commandLine = "%s <options> input_meta_format output_prefix_format\nTakes a time series vector volume (chunked) and splits it for parallel computation.\n";

#define SIZE_OPTION_LIST 8

#define OPTION_START 0
#define OPTION_END 1
#define OPTION_SKIP 2
#define OPTION_META 3
#define OPTION_DATA 4
#define OPTION_MMAP 5
#define OPTION_FILE 6
#define OPTION_MAX 7

char* optionList[] =
  {
    "-start",
    "-end",
    "-skip",
    "-meta",
    "-data",
    "-mmap",
    "-file",
    "-max",
  };

char* optionArg[] =
  {
    "N",
    "N",
    "N",
    "<string>",
    "<string>",
    "",
    "",
    "N",
  };

char* optionHelp[] =
  {
    "start time for output sequence [default = 0]",
    "end time for output sequence (non-inclusive) [default = t (time length of input)]",
    "skip (step size) for output sequence [default = 1]",
    "parallel split metadata output filename suffix [default = .ptmeta]",
    "parallel split data output filename suffix [default = .ptdata]",
    "open input files using mmap instead of file I/O [default = true]",
    "open input files using file I/O instead of mmap [default = false]",
    "maximimum size per output file in bytes [default = 1073741824 (1 GB)]",
  };

#define SIZE_INPUT_LIST 2

char* inputList[] =
  {
    "input_meta_format",
    "output_prefix_format",
  };

char* inputHelp[] =
  {
    "printf format string for input meta filename sequence, requires exactly one number format in the string.  assumes the number ordering is 0+ with no skips.",
    "printf format string for output prefix filename sequence, requires exactly one number format in the string.",
  };

void printHelp(char* argv0) {
  if(mpi_rank == 0) {
    fprintf(stderr, commandLine, argv0);
    
    fprintf(stderr, "\nrequired inputs:\n");
    for(int i = 0; i < SIZE_INPUT_LIST; i = i + 1) {
      fprintf(stderr, "%s: %s\n", inputList[i], inputHelp[i]);
    }
    
    fprintf(stderr, "\noptions:\n");
    for(int i = 0; i < SIZE_OPTION_LIST; i = i + 1) {
      fprintf(stderr, "%s %s : %s\n", optionList[i], 
	      optionArg[i], optionHelp[i]);
    }
    fprintf(stderr, "\n");
    
    fflush(stderr);
  }

  MPI_Finalize();
  exit(1);
}

void printNotEnoughArgs(char* option, char* argv0) {
  if(mpi_rank == 0) {
    fprintf(stderr, "ERROR: not enough arguments to %s\n\n", option);
  }

  printHelp(argv0);
}

CommandArguments* commandParse(int argc, char* argv[])
{
  if(argc < 3) {
    // exit here
    printHelp(argv[0]);
  }

  CommandArguments* args = new CommandArguments();
  setDefaults(args);

  // go through all the arguments except the last two
  for(int i = 1; i < argc - 2; i = i + 1) {
    // start
    if(!strcmp(argv[i], optionList[OPTION_START])) {
      if(i + 1 < argc - 2) {
	args->startFile = atoi(argv[i + 1]);
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_START], argv[0]);
      }
    }
    // end
    else if(!strcmp(argv[i], optionList[OPTION_END])) {
      if(i + 1 < argc - 2) {
	args->endFile = atoi(argv[i + 1]);
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_END], argv[0]);
      }
    }
    // skip
    else if(!strcmp(argv[i], optionList[OPTION_SKIP])) {
      if(i + 1 < argc - 2) {
	args->skipFile = atoi(argv[i + 1]);
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_SKIP], argv[0]);
      }
    }
    // meta
    else if(!strcmp(argv[i], optionList[OPTION_META])) {
      if(i + 1 < argc - 2) {
	args->metaExt = argv[i + 1];
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_META], argv[0]);
      }
    }
    // data
    else if(!strcmp(argv[i], optionList[OPTION_DATA])) {
      if(i + 1 < argc - 2) {
	args->dataExt = argv[i + 1];
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_DATA], argv[0]);
      }
    }
    // mmap
    else if(!strcmp(argv[i], optionList[OPTION_MMAP])) {
      args->useMMap = 1;
    }
    // file
    else if(!strcmp(argv[i], optionList[OPTION_FILE])) {
      args->useMMap = 0;
    }
    // max
    else if(!strcmp(argv[i], optionList[OPTION_MAX])) {
      if(i + 1 < argc - 2) {
	args->maxFile = atoi(argv[i + 1]);
	i = i + 1;
      }
      else {
	printNotEnoughArgs(optionList[OPTION_MAX], argv[0]);
      }
    }
    // everything else
    else {
      fprintf(stderr, "ERROR: unknown argument %s\n\n", argv[i]);
      printHelp(argv[0]);
    }
  }

  args->inputFormat = argv[argc - 2];
  args->outputFormat = argv[argc - 1];

  // sanity check
  if(args->skipFile < 1) {
    if(mpi_rank == 0) {
      fprintf(stderr, "ERROR: file skip needs to be > 0\n\n");
    }
    printHelp(argv[0]);
  }

  if(strcmp(args->metaExt, args->dataExt) == 0) {
    if(mpi_rank == 0) {
      fprintf(stderr, 
	      "ERROR: output meta suffix and data suffix cannot equal\n\n");
    }
    printHelp(argv[0]);
  }

  return args;
}
