#!/usr/bin/python

#/*
# * Copyright (c) 2009, Jonathan L. Woodring and Han-Wei Shen
# * All rights reserved.
# *
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions are met:
# *     * Redistributions of source code must retain the above copyright
# *       notice, this list of conditions and the following disclaimer.
# *     * Redistributions in binary form must reproduce the above copyright
# *       notice, this list of conditions and the following disclaimer in the
# *       documentation and/or other materials provided with the distribution.
# *     * Neither the name of The Ohio State University, Jonathan
# *       L. Woodring, Han-Wei Shen, nor the names of its contributors
# *       may be used to endorse or promote products derived from this
# *       software without specific prior written permission.
# *
# * THIS SOFTWARE IS PROVIDED BY JONATHAN L. WOODRING AND HAN-WEI SHEN
# * ''AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE OHIO
# * STATE UNIVERSITY, JONATHAN L. WOODRING, OR HAN-WEI SHEN BE LIABLE
# * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# */

from os import *
from sys import *
from re import *
from math import *
from cPickle import *
from gzip import *

from sequence_funcs import *

#
# usage print
#
def printHelp(argv):
    stderr.write('usage: %s <options> input_cinfo_format output_filename\n\n' % argv[0])

    stderr.write('required inputs:\n')
    stderr.write('input_cinfo_format: printf format string for input cinfo filename sequence, requires exactly one number format in the string.\n')
    stderr.write('output_filename: filename for the sequence file\n\n')

    stderr.write('options:\n')
    stderr.write('-start N : start time for input sequence [default = 0]\n')
    stderr.write('-end N : end time for input sequence (non-inclusive) [default = 1]\n')
    stderr.write('-skip N : skip time for input sequence [default = 1]\n')
    stderr.write('-cull N : probability edge cull for graph [default = .5\n')
    stderr.write('-power N. : power that the distance are raised [default = 2.0]\n')
    stderr.write('-l2 : use L2 histogram distance [default = false]\n')
    stderr.write('-chi2 : use chi^2 histogram distance [default = false]\n')
    stderr.write('-emd : use Earth Mover\'s distance [default = true]\n')
    stderr.write('-bins N : use N bins for the displayed histograms [default = 200]\n')

    exit()

#
# argument parser
#
def parseArgs(argv):
    if len(argv) < 3:
        printHelp(argv)

    # default values
    data = {}
    data['start'] = 0
    data['end'] = 1
    data['skip'] = 1
    data['cull'] = .5
    data['power'] = 2.0
    data['bins'] = 200

    # go through the arguments
    i = 1
    while i < len(argv) - 2:
        if argv[i] == '-start':
            if i + 1 < len(argv) - 2:
                data['start'] = int(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -start\n\n')
                printHelp(argv)

        elif argv[i] == '-end':
            if i + 1 < len(argv) - 2:
                data['end'] = int(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -end\n\n')
                printHelp(argv)

        elif argv[i] == '-skip':
            if i + 1 < len(argv) - 2:
                data['skip'] = int(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -skip\n\n')
                printHelp(argv)

        elif argv[i] == '-cull':
            if i + 1 < len(argv) - 2:
                data['cull'] = float(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -cull\n\n')
                printHelp(argv)

        elif argv[i] == '-power':
            if i + 1 < len(argv) - 2:
                data['power'] = float(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -power\n\n')
                printHelp(argv)

        elif argv[i] == '-l2':
            DIST.dist = DIST.L2dist

        elif argv[i] == '-chi2':
            DIST.dist = DIST.chi2dist

        elif argv[i] == '-emd':
            DIST.dist = DIST.EMDdist

        elif argv[i] == '-bins':
            if i + 1 < len(argv) - 2:
                data['bins'] = int(argv[i+1])
                i = i + 1
            else:
                stderr.write('ERROR: not enough arguments for -bins\n\n')
                printHelp(argv)

        else:
            stderr.write('ERROR: unknown argument %s\n\n' % argv[i])
            printHelp(argv)

        i = i + 1

    data['format'] = argv[-2]
    outfn = argv[-1]

    # sanity checks
    if data['end'] < data['start']:
        stderr.write('WARNING: end < start, setting end = start + 1')
        data['end'] = data['start'] + 1

    if data['skip'] < 1:
        stderr.write('ERROR: skip value < 1\n\n')
        printHelp(argv)

    if data['bins'] < 1:
        stderr.write('ERROR: bins < 1\n\n')
        printHelp(argv)
        
    return data, outfn


#
# parse the data out of the files
#
def readHistograms(data):
    f = file(data['format'] % data['start'], 'r')

    header = {}

    # create a header description
    header['clustering'] = compile("clusteringtype:\s*(\S*)").search(f.readline()).group(1)
    header['vector'] = compile("vectortype:\s*(\S*)").search(f.readline()).group(1)
    header['filter'] = compile("filtertype:\s*(\S*)").search(f.readline()).group(1)
    header['distance'] = compile("distancetype:\s*(\S*)").search(f.readline()).group(1)
    header['width'] = int(compile("vectorwidth:\s*(\S*)").search(f.readline()).group(1))
    header['bins'] = int(compile("histogrambins:\s*(\S*)").search(f.readline()).group(1))

    timesteps = []

    # read each file and build the time histogram
    for i in range(data['start'], data['end'], data['skip']):
        f = file(data['format'] % i, 'r')

        desc = {}

        # skip the starting stuff
        for i in range(0, 6):
            f.readline()

        # read the information about this timestep
        desc['min'] = float(compile("min:\s*(\S*)").search(f.readline()).group(1))
        desc['max'] = float(compile("max:\s*(\S*)").search(f.readline()).group(1))
        desc['time'] = int(compile("timestep:\s*(\S*)").search(f.readline()).group(1))
        desc['number'] = int(compile("clusters:\s*(\S*)").search(f.readline()).group(1))
        desc['iter'] = int(compile("iterations:\s*(\S*)").search(f.readline()).group(1))
        desc['error'] = float(compile("averageerror:\s*(\S*)").search(f.readline()).group(1))

        clusters = []

        for j in range(0, desc['number']):
            info = {}

            # parse the clusters
            info['centroid'] = map(float, compile("cluster:(.*)").search(f.readline()).group(1).split())
            info['error'] = float(compile("error:\s*(\S*)").search(f.readline()).group(1))
            info['count'] = int(compile("count:\s*(\S*)").search(f.readline()).group(1))

            histogram = []
            for k in range(0, header['width']):
                histogram.append(map(int, compile("timehistogram:(.*)").search(f.readline()).group(1).split()))

            info['timehistogram'] = histogram
            clusters.append(info)

        desc['clusters'] = clusters
        timesteps.append(desc)

    # return the header and the time histograms
    return header, timesteps

#
# main, do everything
#

data, outfn = parseArgs(argv)
header, timesteps = readHistograms(data)
header, timesteps = adjustDisplayedHistograms(header, timesteps, data['bins'])
graph, weights = generateProbs(timesteps, data['power'], header)
starters, enders, isolated = cullEdges(graph, weights, data['cull'], len(timesteps))
paths = removeDuplicates(pathFind(graph, starters, enders, len(timesteps)))
metrics = calcMetrics(weights, paths, len(timesteps))

f = GzipFile(outfn, 'wb')
p = Pickler(f)

p.dump(data)
p.dump(header)
p.dump(timesteps)
p.dump(graph)
p.dump(weights)
p.dump(paths)
p.dump(metrics)

f.close()

