#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

FILE* openVolume(char* fn, size_t x, size_t y, size_t z)
{
  FILE* fp = fopen(fn, "r");

  if(!fp) {
    fprintf(stderr, "unable to open %s for reading\n", fn);
    exit(1);
  }

  fseek(fp, 0, SEEK_END);
  
  if(x * y * z > ftell(fp) / sizeof(float)) {
    fprintf(stderr, "dimensions larger than %s file size\n", fn);
    exit(1);
  }

  rewind(fp);
  return fp;
}

void simpleHSV(float data, float bias, float scale, float* color) {
  // between red and blue
  float h = (data - bias) * scale * 4.0;
  h = h < 0 ? 0 : (h > 4.0 ? 4.0 : h);

  float b;
  float f = modff(h, &b);
  
  float s = 1.0;
  float v = 1.0;

  float p = v * (1 - s);
  float q = v * (1 - s * f);
  float r = v * (1 - s * (1 - f));

  // actually the color won't go over blue (h <= 4) but this is generalized
  // HSV to RGB conversion
  switch((int)b) {
  case 0: {
    color[0] = v;
    color[1] = r;
    color[2] = p;
  } break;

  case 1: {
    color[0] = q;
    color[1] = v;
    color[2] = p;
  } break;

  case 2: {
    color[0] = p;
    color[1] = v;
    color[2] = r;
  } break;

  case 3: {
    color[0] = p;
    color[1] = q;
    color[2] = v;
  } break;

  case 4: {
    color[0] = r;
    color[1] = p;
    color[2] = v;
  } break;

  default: {
    color[0] = v;
    color[1] = p;
    color[2] = q;
  } break;
  }
}

// a triangle hat
float simpleRamp(float value, float mid, float tail) {
  if(value <  mid - tail || value > mid + tail) {
    return 0.0;
  }
  
  return 1.0 - (fabs(value - mid) / tail);
}

void premultiply(float* color) {
  color[0] = color[0] * color[3];
  color[1] = color[1] * color[3];
  color[2] = color[2] * color[3];
}

void additiveOp(float* color0, float* color1, float* out) {
  out[0] = color0[0] + color1[0];
  out[1] = color0[1] + color1[1];
  out[2] = color0[2] + color1[2];
  out[3] = color0[3] + color1[3];
}

void frontToBackOp(float* front, float* back, float* out) {
  float scale = 1 - front[3];
  out[0] = front[0] + back[0] * scale;
  out[1] = front[1] + back[1] * scale;
  out[2] = front[2] + back[2] * scale;
  out[3] = front[3] + back[3] * scale;
}  

void writeRGBA(FILE* output, float* color) {
  color[0] = color[0] < 0 ? 0 : (color[0] > 1 ? 1 : color[0]);
  color[1] = color[1] < 0 ? 0 : (color[1] > 1 ? 1 : color[1]);
  color[2] = color[2] < 0 ? 0 : (color[2] > 1 ? 1 : color[2]);
  color[3] = color[3] < 0 ? 0 : (color[3] > 1 ? 1 : color[3]);

  unsigned char conv[4];
  conv[0] = (unsigned char)(color[0] < 1.0 ? color[0] * 256 : 255);
  conv[1] = (unsigned char)(color[1] < 1.0 ? color[1] * 256 : 255);
  conv[2] = (unsigned char)(color[2] < 1.0 ? color[2] * 256 : 255);
  conv[3] = (unsigned char)(color[3] < 1.0 ? color[3] * 256 : 255);

  fwrite(conv, sizeof(unsigned char), 4, output);
}

void additiveFix(float* color) {
  if(color[3] > 1) {
    color[0] = color[0] / color[3];
    color[1] = color[1] / color[3];
    color[2] = color[2] / color[3];
    color[3] = 1.0;
  }
}

int main(int argc, char* argv[]) {
  // read command line arguments
  if(argc < 7) {
    fprintf(stderr, "usage: %s <x> <y> <z> <min float> <max float> <list of float volume filenames>\n",
	    argv[0]);
    exit(1);
  }

  // get the dimensions of the float blocks
  size_t x = atol(argv[1]);
  size_t y = atol(argv[2]);
  size_t z = atol(argv[3]);

  float minv = atof(argv[4]);
  float maxv = atof(argv[5]);
  float biasv = minv;
  float scalev = 1.0 / (maxv - minv);
  float midv = (minv + maxv) * .5;
  float tailv = (maxv - minv) * .5;

  size_t numinputs = argc - 6;

  float mint = 0.0;
  float maxt = numinputs - 1;
  float biast = mint;
  float scalet = 1.0 / (maxt - mint);
  float midt = (mint + maxt) * .5;
  float tailt = (maxt - mint) * .5;
  
  // open the files and check them
  FILE* inputs[numinputs];

  for(size_t i = 0; i < numinputs; i = i + 1) {
    inputs[i] = openVolume(argv[i + 6], x, y, z);
  }

  // some different example operations to be written
  // additive, compositing, max time value, and intersection

  // they are written out as color volumes (not scalar)
  // you will need a color volume renderer to view them

  // the "tricky" part is the time ordinal->color mapping
  // which is why the color volumes are written, instead of scalar volumes
  
  // the reason is:
  // while some of these operations use post-classification (operate then color
  // e.g., maxtime, intersection), the data could be written as
  // time ordinal (scalar) volumes, and use a mapping in the volume
  // renderer: but things that use pre-classification (color then operate
  // e.g., additive, compositing) the result is color, and thus color
  // volumes are written

  // Utah's ImageVis3D sort of works for rendering color volumes
  // (have to turn on isosurface rendering mode, 
  // but it doesn't respect the alpha value for rendering)
  FILE* additiveOutput = fopen("additive.rgba", "w");
  FILE* compositeOutput = fopen("composite.rgba", "w");
  FILE* maxTimeOutput = fopen("maxtime.rgba", "w");
  FILE* inOutput = fopen("intersection.rgba", "w");

  // write qvis .dat headers (this color volume format can be read by Utah's
  // ImageVis3D volume renderer)
  FILE* header;

  header = fopen("additive.dat", "w");
  fprintf(header, "ObjectFileName: additive.rgba\n");
  fprintf(header, "TaggedFileName: ---\n");
  fprintf(header, "Resolution: %ld %ld %ld\n", x, y, z);
  fprintf(header, "SliceThickness: 1 1 1\n");
  fprintf(header, "Format: UCHAR4\n");
  fprintf(header, "ObjectType: TEXTURE_VOLUME_OBJECT\n");
  fprintf(header, "ObjectModel: RGBA\n");
  fprintf(header, "GridType: EQUIDISTANT\n");
  fclose(header);

  header = fopen("composite.dat", "w");
  fprintf(header, "ObjectFileName: composite.rgba\n");
  fprintf(header, "TaggedFileName: ---\n");
  fprintf(header, "Resolution: %ld %ld %ld\n", x, y, z);
  fprintf(header, "SliceThickness: 1 1 1\n");
  fprintf(header, "Format: UCHAR4\n");
  fprintf(header, "ObjectType: TEXTURE_VOLUME_OBJECT\n");
  fprintf(header, "ObjectModel: RGBA\n");
  fprintf(header, "GridType: EQUIDISTANT\n");
  fclose(header);

  header = fopen("maxtime.dat", "w");
  fprintf(header, "ObjectFileName: maxtime.rgba\n");
  fprintf(header, "TaggedFileName: ---\n");
  fprintf(header, "Resolution: %ld %ld %ld\n", x, y, z);
  fprintf(header, "SliceThickness: 1 1 1\n");
  fprintf(header, "Format: UCHAR4\n");
  fprintf(header, "ObjectType: TEXTURE_VOLUME_OBJECT\n");
  fprintf(header, "ObjectModel: RGBA\n");
  fprintf(header, "GridType: EQUIDISTANT\n");
  fclose(header);

  header = fopen("intersection.dat", "w");
  fprintf(header, "ObjectFileName: intersection.rgba\n");
  fprintf(header, "TaggedFileName: ---\n");
  fprintf(header, "Resolution: %ld %ld %ld\n", x, y, z);
  fprintf(header, "SliceThickness: 1 1 1\n");
  fprintf(header, "Format: UCHAR4\n");
  fprintf(header, "ObjectType: TEXTURE_VOLUME_OBJECT\n");
  fprintf(header, "ObjectModel: RGBA\n");
  fprintf(header, "GridType: EQUIDISTANT\n");
  fclose(header);

  // here is the meat

  // for all points, apply comparison operation
  // think "fragment shader" or "kernel operation" or "data parallelism"
  // these can be easily written as Cuda, GLSL, or OpenCL programs
  for(size_t k = 0; k < z; k = k + 1) {
    for(size_t j = 0; j < y; j = j + 1) {
      for(size_t i = 0; i < x; i = i + 1) {
	float addcolor[4] = {0.0, 0.0, 0.0, 0.0};
	float compcolor[4] = {0.0, 0.0, 0.0, 0.0};

	float maxvalue = FLT_MIN;
	float maxtime = 0.0; 

	float in = 1.0;
	float insum = 0.0;

	// op applied over time (chronovolumes)

	// if this was arbitrary composition, instead of time operations
	// (chronovolumes), this would be an operation tree to apply 
	// operations on selected inputs in composition order
	for(size_t t = 0; t < numinputs; t = t + 1) {
	  float color[4];
	  float value;

	  // read/get value
	  // you can put some neighborhood sampling
	  // filters in here to smooth data when applying operations
	  fread(&value, sizeof(float), 1, inputs[t]);


	  // pre-classified color lookup
	  // lookup color by time value
	  simpleHSV(t, biast, scalet, color);
	  // alpha by data value
	  color[3] = simpleRamp(value, midv, tailv);
	  // premultiplied color
	  premultiply(color);


	  // the operations
	  // pre-classified additive operation (add colors over time)
	  additiveOp(addcolor, color, addcolor);

	  // pre-classified composite operation (composite colors over time)
	  frontToBackOp(compcolor, color, compcolor);

	  // max time value operation
	  // (find the time step that has the maximum value)
	  maxtime = value > maxvalue ? t : maxtime;
	  maxvalue = value > maxvalue ? value : maxvalue;

	  // in (intersection) operation using normalized value > 0.5 
	  // as the threshold for determining the sets
	  // (find the values that intersect in space over time)
	  in = in * (value > 0.5 ? 1.0 : 0.0);
	  insum = insum + value;
	}

	// post-classified color for max time value operation
	float maxtcolor[4] = {0.0, 0.0, 0.0, 0.0};
	// lookup color by max time value
	simpleHSV(maxtime, biast, scalet, maxtcolor);
	// alpha by max value
	maxtcolor[3] = simpleRamp(maxvalue, midv, tailv);
	// premultiplied color
	premultiply(maxtcolor);
	
	// post-classified color for in (intersection) operation
	float incolor[4] = {0.0, 0.0, 0.0, 0.0};
	if(in > 0) {
	  insum = insum / numinputs;
	  // lookup color by average value
	  simpleHSV(insum, biasv, scalev, incolor);
	  // alpha by average value
	  incolor[3] = simpleRamp(insum, midv, tailv);
	  // premultiplied color
	  premultiply(incolor);
	}

	// fixup the additive color, if the alpha overflows
	additiveFix(addcolor);

	// write the result for each point as a uchar RGBA vector
	// (color volumes)
	writeRGBA(additiveOutput, addcolor);
	writeRGBA(compositeOutput, compcolor);
	writeRGBA(maxTimeOutput, maxtcolor);
	writeRGBA(inOutput, incolor);
      }
    }
  }

  // close files
  fclose(additiveOutput);
  fclose(compositeOutput);
  fclose(maxTimeOutput);
  fclose(inOutput);
}
